/*
 * (s)ed for text processing input file as a one buffer
 * (without line by line)
 *
 * Copyright (C) Vladimir Oleynik <dzo@simtreas.ru> 2008-2023
 * Licensed under GPLv2 or later, see file LICENSE in this tarball for details.
*/

#include "opt_gen.h"
BITS_GEN(DECONVERT, NEED_ALLOC, NEED_ZTERM, NEED_FREE)
#include "opt_gen.h"

#include "bsed.h"
#include <time.h>
#include <getopt.h>

const char *prg_name;

static void usage(void) __attribute__ ((__noreturn__));
static void usage(void)
{
	fprintf(stderr, ": %s [-e script]* [-f script-file]* [-i file] [files]*\n", prg_name);
	exit(2);
}

static struct REPLACES *replaces, *tail;
static unsigned char uused[256];
static size_t bufsz, idx, la, lb2;
static char *m0, *space, *buf2;
static int flg_need_unused;
static const char *input_file;
static char *prg_buf;
static long count;
static unsigned int allow_replace;

#define MAX_OFFS 32
static regmatch_t rm[MAX_OFFS];

static void replace1c_buf(int f, unsigned int c, char *b, size_t l)
{
	for(; l != 0; ) {
		char *p = memchr(b, f, l);
		if(p == NULL) break;
		*p++ = c;
		l -= (p - b);
		b = p;
	}
}

static void printb(FILE *out, char *b, size_t l, struct REPLACES *rs)
{
	if(used[0] != '\0')
		replace1c_buf(used[0], '\0', b, l);
	if(fwrite(b, 1, l, out) != l)
		error_msg(1, "%zu:  /: %m", rs->line);
	fflush(out);
	if(used[0] != '\0')
		replace1c_buf(0, used[0], b, l);
}


static char *deconvert(char *buf, size_t sz, int flgs)
{
	if((flgs & (DECONVERT_NEED_ALLOC|DECONVERT_NEED_ZTERM))) {
		char *b;
		size_t need_alloc = sz;

		if((flgs & DECONVERT_NEED_ZTERM))
			need_alloc++;
		if((flgs & DECONVERT_NEED_ALLOC))
			b = memcpy(xmalloc(need_alloc), buf, sz);
		else
			b = xrealloc(buf, need_alloc);
		if((flgs & DECONVERT_NEED_ZTERM))
			b[sz] = '\0';
		if((flgs & DECONVERT_NEED_FREE))
			free(buf);
		buf = b;
	}
	replace1c_buf(used[0], '\0', buf, sz);
	return buf;
}

static char *load_file(const char *f, char *buf, const char *last_used, char *i)
{
	if(f == NULL) {
		size_t l = strlen(last_used) + 1;
		buf = xrealloc(buf, idx + l);
		memcpy(buf + idx, last_used, l);
		idx += l - 1;
	} else {
		int ch;
		FILE *file;

		if(f[0] == '-' && f[1] == '\0' && i == NULL) {
			f = "<stdin>";
			file = stdin;
		} else {
			file = fopen(f, "r");
			if(file == NULL)
				error_msg(1, "  '%s'  : %m", f);
		}
		if(last_used != NULL)
			input_file = f;
		while ((ch = getc(file)) != EOF) {
			/* grow the buffer as necessary */
			if (idx == bufsz)
				buf = xrealloc(buf, bufsz += 4082);
			if(last_used != NULL)
				used[(size_t)ch] = 1;
			buf[idx++] = (char)ch;
		}
		if (ferror(file))
			error_msg(1, "  '%s'", f);
		if(file != stdin)
			fclose(file);
	}
	return buf;
}

static void xregcomp(regex_t *preg, const char *pattern_name, struct REPLACES *rs, const char *regex, size_t sz)
{
	int ret = REG_EXTENDED;
	if((rs->mods & BIT_IGNORE_CASE))
		ret |= REG_ICASE;

	if ((ret = tre_regncomp(preg, regex, sz, ret)) != 0) {
		size_t errmsgsz = regerror(ret, preg, NULL, 0);
		char *errmsg = xmalloc(errmsgsz);
		regerror(ret, preg, errmsg, errmsgsz);
		error_msg(4, "%zu:   %s: %s",
				rs->line, pattern_name, errmsg);
	}
}

static void tst_allow_subexpr(struct REPLACES *rs)
{
	if((allow_replace & BIT_INPUT_STORED)) return;
	error_msg(3, "%zu: %c -   \\h or \\@...\\Z or \\z\n\
         ", rs->line, rs->opt);
}

static char *set_repl(struct REPLACES *rs, size_t *pl, char **store)
{
	size_t l = 0, nr, of, diff, lc = 0, lst = 0, lul, *need_convert = NULL, *lp = NULL;
	const char *replace = rs->find, *r = replace;
	char *rp, *r0, cbuf[64], stbuf[64];
	size_t fil = 0;
	unsigned int fast_repl = rs->mods & BIT_FAST_REPL;
	int flg_up_low = 0;

	if(rs->n_offs) {
		need_convert = xmalloc(rs->n_offs * sizeof(size_t) * 2);
		lp = need_convert + rs->n_offs;
	}
	for(nr = 0; nr < rs->n_offs; nr++) {
		diff = ((replace + rs->offs[nr]) - r);
		l += diff;
		r += diff;
		of = rs->codes[nr];
		if(of == CODE_UNUSED) {
			continue;
		} else if(of == CODE_FIND_STATUS) {
			lst = sprintf(stbuf, "%u", allow_replace & BIT_FOUND);
			l += lst;
			continue;
		} else if(of == CODE_UPPER || of == CODE_LOWER) {
			flg_up_low = of;
			continue;
		} else if(of == CODE_COUNT) {
			lc = sprintf(cbuf, "%ld", count);
			l += lc;
			continue;
		} else if(of == CODE_INPUT_FILE) {
			fil = strlen(input_file);
			l += lp[nr] = calc_up_low(input_file, fil, flg_up_low, &need_convert[nr]);
		} else if(of == CODE_INB) {
			l += lp[nr] = calc_up_low(m0, idx, flg_up_low, &need_convert[nr]);
		} else if(of == CODE_OUT) {
			l += lp[nr] = calc_up_low(space, la, flg_up_low, &need_convert[nr]);
		} else if(of == CODE_BUF2) {
			l += lp[nr] = calc_up_low(buf2, lb2, flg_up_low, &need_convert[nr]);
		} else if(of == CODE_HEAD) {
			if(fast_repl && l > (size_t)rm[0].rm_so)
				fast_repl = 0;
			lul = rm[0].rm_so;
			of = 0;
		       l_calc_sz:
			tst_allow_subexpr(rs);
			l += lp[nr] = calc_up_low(m0 + of, lul, flg_up_low, &need_convert[nr]);
		} else if(of == CODE_END) {
			if(fast_repl && l > (size_t)rm[0].rm_eo)
				fast_repl = 0;
			lul = idx - rm[0].rm_eo;
			of = rm[0].rm_eo;
			goto l_calc_sz;
		} else if(rm[of].rm_so != -1) {
			if(fast_repl && l > (size_t)rm[of].rm_so)
				fast_repl = 0;
			lul = rm[of].rm_eo - rm[of].rm_so;
			of = rm[of].rm_so;
			goto l_calc_sz;
		}
		flg_up_low = 0;
	}
	l += rs->lf - (r - replace);
	r = replace;
	if(fast_repl) {
		if(l > idx)
			m0 = xrealloc(m0, l);
		r0 = rp = m0;
	} else {
		if((rs->mods & BIT_INS_REPL))
			r0 = rp = xmalloc(idx + l);
		else
			r0 = rp = xmalloc(l);
	}
	if((rs->mods & BIT_H_MOD))
		count = 0;
	for(nr = 0; nr < rs->n_offs; nr++) {
		diff = ((replace + rs->offs[nr]) - r);
		rp = memcpy(rp, r, diff) + diff;
		r += diff;
		of = rs->codes[nr];
		if(of == CODE_UNUSED) {
			continue;
		} else if(of == CODE_FIND_STATUS) {
			rp = memcpy(rp, stbuf, lst) + lst;
			continue;
		} else if(of == CODE_UPPER || of == CODE_LOWER) {
			flg_up_low = of;
			continue;
		} else if(of == CODE_COUNT) {
			rp = memcpy(rp, cbuf, lc) + lc;
			continue;
		} else if(of == CODE_INPUT_FILE) {
			up_low(input_file, fil, flg_up_low, rp, need_convert[nr], 1);
		} else if(of == CODE_OUT) {
			up_low(space, la, flg_up_low, rp, need_convert[nr], 1);
		} else if(of == CODE_BUF2) {
			up_low(buf2, lb2, flg_up_low, rp, need_convert[nr], 1);
		} else if(of == CODE_INB) {
			lul = idx;
			of = 0;
		       lrepl:
			if(!up_low(m0 + of, lul, flg_up_low, rp, need_convert[nr], 0)) {
				if(fast_repl) {
					if(rp != (m0+of))
						memmove(rp, m0 + of, lul);
				} else {
					if(r0 != m0)
						memcpy(rp, m0 + of, lul);
					else
						memmove(rp, m0 + of, lul);
				}
			}
		} else if(of == CODE_HEAD) {
			lul = rm[0].rm_so;
			of = 0;
			goto lrepl;
		} else if(of == CODE_END) {
			lul = idx - rm[0].rm_eo;
			of = rm[0].rm_eo;
			goto lrepl;
		} else if(rm[of].rm_so != -1) {
			lul = rm[of].rm_eo - rm[of].rm_so;
			of = rm[of].rm_so;
			goto lrepl;
		}
		rp += lp[nr];
		flg_up_low = 0;
		if((rs->mods & BIT_H_MOD))
			count += need_convert[nr];
	}
	free(need_convert);
	memcpy(rp, r, rs->lf - (r - replace));

	if((rs->mods & BIT_ADD_REPL)) {
		*store = xrealloc(*store, *pl + l);
		memcpy(*store + *pl, r0, l);
		*pl += l;
		free(r0);
		r0 = m0;
	} else if((rs->mods & BIT_INS_REPL)) {
		memcpy(r0 + l, m0, idx);
		idx += l;
		free(m0);
		m0 = r0;
	} else if(fast_repl) {
		if(l < idx)
			m0 = xrealloc(m0, l);
		idx = l;
		r0 = m0;
	} else {
		*pl = l;
		if(store != NULL) {
			free(*store);
			*store = r0;
		}
	}
	return r0;
}

static char *add_cmd(char cmd, char *o, unsigned int mods, size_t *pline)
{
	const char *s = o;
	char *p = NULL;
	unsigned char *codes = NULL;
	size_t lf = 0, *offs = NULL, n_offs = 0, start_line = *pline;
	struct REPLACES *rp;
	regex_t *R = NULL;

	if((mods & BIT_CMD_IS_COMPLEX)) {
	    size_t line = start_line, l_flg_need_unused = 0;
	    char *e, delim = *s++;
	    char flg_up_low = 0;

	    if(delim == '\0')
		error_msg(2, "%zu:    -", line);
	    if(delim == '\\')
		error_msg(2, "%zu:  - '\\'", line);
	    e = p = o;

	    offs = xmalloc(sizeof(*offs) * (strlen(s) / 2 + 2));
	    codes = xmalloc(sizeof(*codes) * (strlen(s) / 2 + 2));
	    if((mods & BIT_ADD_HEAD)) {
		offs[0] = 0;
		codes[0] = CODE_HEAD;
		n_offs = 1;
	    }

	    while (*s) {
		if (*s == '\\') {
			s++;
			if(*s == '\0')
				error_msg(2, "%zu:  ", line);

#define SET_ESC_POS_CODE(code, tst_up_low) do { \
				offs[n_offs] = (e - p); \
				codes[n_offs++] = code; \
				if(!tst_up_low && flg_up_low) \
					error_msg(2, "%zu:    \\%c", line, flg_up_low); \
				flg_up_low = 0; \
				s++; \
			       } while(0)

			if(*s >= '@' && *s <= 'Z') {
				int of = (*s - '@');
				SET_ESC_POS_CODE(of, 1);
			} else if(*s == 'y') {
				e++;    /* any */
				l_flg_need_unused++;
				flg_need_unused = 1;
				SET_ESC_POS_CODE(CODE_UNUSED, 0);
				mods |= BIT_HAVE_UNUSED;
			} else if(*s == '~') SET_ESC_POS_CODE(CODE_FIND_STATUS, 0);
			  else if(*s == 'o') SET_ESC_POS_CODE(CODE_INPUT_FILE, 1);
			  else if(*s == 'i') SET_ESC_POS_CODE(CODE_INB, 1);
			  else if(*s == 'h') SET_ESC_POS_CODE(CODE_HEAD, 1);
			  else if(*s == 'p') SET_ESC_POS_CODE(CODE_OUT, 1);
			  else if(*s == 'g') SET_ESC_POS_CODE(CODE_BUF2, 1);
			  else if(*s == 'c') SET_ESC_POS_CODE(CODE_COUNT, 0);
			  else if(*s == 'z') SET_ESC_POS_CODE(CODE_END, 1);
			  else if(*s == 'u') {
				SET_ESC_POS_CODE(CODE_UPPER, 0);
				flg_up_low = *s;
			 } else if(*s == 'l') {
				SET_ESC_POS_CODE(CODE_LOWER, 0);
				flg_up_low = *s;
			 } else if(*s == '\n') {
				line++;
				s++;
			} else {
				char c;
				if(flg_up_low)
					error_msg(2, "%zu:    \\%c", line, flg_up_low);
				c = *s == delim ? *s++ : process_escape_sequence(&s);
				*e++ = c;
				uused[(unsigned char)c] = 1;
			}
		} else {
			if(flg_up_low)
				error_msg(2, "%zu:    \\%c", line, flg_up_low);
			if(*s == delim)
				break;
			if(*s == '\n')
				line++;
			uused[(unsigned char)*s] = 1;
			*e++ = *s++;
		}
	    }
	    if(*s == '\0')
		error_msg(2, "%zu:    -", start_line);

	    lf = (e - p);

	    if((mods & BIT_REGEX)) {
		if(!lf && !n_offs)
			error_msg(2, "%zu:     ", line);
		if((n_offs-l_flg_need_unused))
			mods |= BIT_DYN_REGEX;
		R = xmalloc(sizeof(*R));
	    }
	    *pline = line;
	    if(!n_offs) {
		free(codes);
		codes = NULL;
		free(offs);
		offs = NULL;
	    }
	}

	rp = xmalloc(sizeof(struct REPLACES));
	if(tail == NULL)
		replaces = rp;
	else
		tail->next = rp;
	tail = rp;
	rp->find = p == NULL ? o : p;
	rp->R = R;
	rp->lf = lf;
	rp->offs = offs;
	rp->n_offs = n_offs;
	rp->codes = codes;
	rp->opt = cmd;
	rp->mods = mods;
	rp->line = start_line;
	rp->next = NULL;
	return (char *)s;
}

static char *is_space_cr_inc(char *s, void *p)
{
	size_t *pline = (p == NULL || p == s) ? NULL : p;
	while(*s) {
		if(*s == ' ' || *s == '\r' || *s == '\f' || *s == '\t') {
			if(pline == NULL) return s;
		} else if(*s == '\n') {
			if(pline == NULL) return s;
			(*pline)++;
		} else if(*s == '#') {
			if(pline == NULL) return s;
			while(*++s != '\n') ;
			(*pline)++;
		} else {
			if(p == NULL) return NULL;
			if(pline != NULL) return s;
		}
		s++;
	}
	return s;
}

static void compile_script(void)
{
	char *s, *e, *pcmd, cmd;
	size_t line = 1, eline;
	unsigned int mods, cmd_R;
	struct REPLACES *rs;
	static const char     condition[]="jtFGgLlv?:";
	static const char rev_condition[]="jFtlLgG?v";
	static const char   complex_cmd[]="IoRcianCAPMNBVOHwsekWTQmz";
	size_t label_level = 0, n_label_stack = 0;
	struct REPLACES **label_stack = NULL;

	for(s = is_space_cr_inc(prg_buf, &line); *s != '\0'; s = is_space_cr_inc(s, &line)) {
		if(strchr("dDxXpEkKSY", *s) != NULL) {
			add_cmd(*s++, NULL, 0, &line);
			continue;
		}

		pcmd = s;
		cmd = *s++;
		if((e = strchr(condition, cmd)) != NULL) {
			s = is_space_cr_inc(s, &line);
			if(*s == '\0')
				error_msg(2, "%zu:   ,  '%c'   ", line, cmd);
			if(*s == '{') {
				if(cmd == 'j' || cmd == ':')
					error_msg(2, "%zu:  ,   '%c' ", line, *s);
				add_cmd(rev_condition[(e-condition)], NULL, BIT_SAVE_FLAGS, &line);
				/* push_label_level */
				if(label_level == n_label_stack)
					label_stack = xrealloc(label_stack, (++n_label_stack) * sizeof(*label_stack));
				label_stack[label_level++] = tail;
				s++;
				continue;
			}
			if(*s == '}')
				error_msg(2, "%zu:  ,   '%c' ", line, *s);
			e = is_space_cr_inc(s, s);
			eline = (e - s);
			s = memmove(pcmd, s, eline);
			pcmd[eline] = '\0';
			if(cmd == ':') {
				for(rs = replaces; rs != NULL; rs = rs->next) {
					if(rs->opt == ':' && strcmp(rs->find, s) == 0)
						error_msg(2, "%zu:    '%s'", line, s);
				}
			}
			add_cmd(cmd, s, (cmd == ':' ? 0 : BIT_IS_LABEL), &line);
			s = e;
			continue;
		}

		if(cmd == '}') {
			/* pop_label_level */
			if(!label_level)
				error_msg(2, "%zu:     {  }", line);
			add_cmd(':', "{}", BIT_SAVE_FLAGS, &line);
			label_stack[--label_level]->find = tail;
			continue;
		}

		cmd = '\0';
		cmd_R = BIT_CMD_IS_COMPLEX;
		rs = NULL;
		eline = line;
		/* set e=/[regex/][replace/]* */
		for(e = s; strchr(complex_cmd, *e) != NULL; e++) ;
		e = is_space_cr_inc(e, &eline);
		s--;

		while(*s != '\0' && strchr(complex_cmd, *s) != NULL) {
			mods = cmd_R;
			if(*s == 'I' || *s == 'o' || *s == 'H' || *s == 'k')
				error_msg(2, "%zu:  '%c'    ", line, *s);
			if(*s == 'a' || *s == 'A')
				mods |= BIT_ADD_REPL;
			else if(*s == 'i')
				mods |= BIT_INS_REPL;
			else if(*s == 'c')
				mods |= BIT_FAST_REPL;
			else if(*s == 'z')
				cmd_R |= BIT_IS_RTEST;
			else if(*s == 'R') {
				mods |= BIT_REGEX;
				cmd_R |= BIT_IS_RTEST;
			}
			cmd = *s;
			for(;;) {
				s++;
				if(*s == 'I') {
					if(cmd != 'R' && cmd != 's')
						error_msg(2, "%zu:   '%c'   '%c'", line, *s, cmd);
					mods |= BIT_IGNORE_CASE;
				} else if(*s == 'o') {
					if(cmd != 'R' && cmd != 'A' && cmd != 'C')
						error_msg(2, "%zu:   '%c'   '%c'", line, *s, cmd);
					if(cmd == 'R') {
						if(rs != NULL)
							error_msg(2, "%zu:  '%c'      ", line, *s);
						add_cmd(':', "{}", 0, &line);
						rs = tail;
						mods |= BIT_IS_LOOP;
					} else {
						mods |= BIT_CC_P_MOD;
					}
				} else if(*s == 'H') {
					if(strchr("aciACPsw", cmd) == NULL)
						error_msg(2, "%zu:   '%c'   '%c'", line, *s, cmd);
					mods |= BIT_H_MOD;
				} else if(*s == 'k') {
					if(cmd != 'P')
						error_msg(2, "%zu:   '%c'   '%c'", line, *s, cmd);
					mods |= BIT_IS_PK;
				} else {
					break;
				}
			}
			if(cmd == 's') {
				struct REPLACES *rss, *resto = NULL;

				if((mods & BIT_IS_RTEST)) {
					add_cmd('F', NULL, BIT_SAVE_FLAGS, &line);
					resto = tail;
				}
				add_cmd('k', NULL, 0, &eline);
				add_cmd(':', "{}", 0, &eline);
				rss = tail;
				e = add_cmd('R', e, BIT_CMD_IS_COMPLEX | BIT_REGEX | (mods & BIT_IGNORE_CASE), &eline);
				e = add_cmd('P', e, BIT_CMD_IS_COMPLEX | BIT_ADD_REPL | BIT_ADD_HEAD | BIT_IS_RTEST | (mods & BIT_H_MOD), &eline);
				add_cmd('c', "/\\z/", BIT_CMD_IS_COMPLEX | BIT_FAST_REPL | BIT_IS_RTEST, &eline);
				add_cmd('t', NULL, 0, &eline);
				tail->find = rss;
				add_cmd('Z', NULL, 0, &eline);
				add_cmd('K', NULL, 0, &eline);
				if((mods & BIT_IS_RTEST)) {
					add_cmd(':', "{}", BIT_SAVE_FLAGS, &eline);
					tail->find = resto;
					resto->find = tail;
				}
			} else {
				e = add_cmd(cmd, e, mods, &eline);
			}
		}

		if(cmd != '\0') {
			s = e + 1;
			line = eline;
			if(rs != NULL) {
				/*   ,    ,
				      , 
				        */
				add_cmd('t', NULL, BIT_IS_LOOP, &line);
				tail->find = rs;
			}
		} else
			error_msg(2, "%zu:   '%c'", line, *s);
	}
	if(replaces == NULL)
		error_msg(2, " ");
	if(label_level)
		error_msg(2, "  %zu  }", label_level);

	for(rs = replaces; rs != NULL; rs = rs->next) {
		if(rs->mods == BIT_IS_LABEL) {
			struct REPLACES *q;
			for(q = replaces; q != NULL; q = q->next) {
				if(q->opt == ':' && strcmp(rs->find, q->find) == 0)
					break;
			}
			rs->find = q != NULL ? q : tail;
		}
	}
}

static long count_strtol(char *s, char **e, struct REPLACES *rs)
{
	long lc;
	char **e0 = e;
	char *pe;

	if(e == NULL)
		e = &pe;
	*e = s;
	errno = 0;
	lc = strtol(s, e, 10);
	if(errno || *e == s || (e0 == NULL && **e != '\0'))
		error_msg(3, "%zu:       ", rs->line);
	return lc;
}

static size_t stack_level;
static char **stack;
static size_t *stack_bsz;

static void push_stack_level(void)
{
	static size_t n_stack;

	if(stack_level == n_stack) {
		stack = xrealloc(stack, (++n_stack) * sizeof(*stack));
		stack_bsz = xrealloc(stack_bsz, n_stack * sizeof(*stack_bsz));
	}
	stack[stack_level] = buf2;
	stack_bsz[stack_level] = lb2;
	stack_level++;
	buf2 = NULL;
	lb2 = 0;
}

static void pop_stack_level(size_t line, char **out, size_t *pl)
{
	if(!stack_level)
		error_msg(2, "%zu: K  ", line);
	free(*out);
	*out = stack[--stack_level];
	*pl = stack_bsz[stack_level];
}

static FILE *sout;     /* musl libc need */

static void initialize(char *a0)
{
	char *a;

	prg_name = a0;
	for (a = a0; *a; )
		if (*(a++) == '/') prg_name = a;
	test_is_utf8();
	sout = stdout;
}

static void to_width(struct REPLACES *rs)
{
	char *e = NULL, *x, *x0;
	long lc = -1;
	size_t l;

	x0 = set_repl(rs, &l, NULL);
	x = memchr(x0, ':', l);
	if(x != NULL)
		lc = count_strtol(x0, &e, rs);
	if(e == NULL || *e != ':' || lc <= 0)
		error_msg(1, "%zu:   ", rs->line);
	x = e + 1;
	l -= (x - x0);
	e = x + l;
	free(space);
	space = xmalloc(l);
	la = 0;
	while(x < e) {
		char *s, *q, *t;
		size_t sz;

		q = memchr(x, '\n', e-x);
		s = x;
		if(q == NULL) {
			/*     \n */
			sz = (e - x);
			x = e;
		} else {
			sz = (q - x);
			x = q + 1;
		}
		if(mbchar_len(s, sz) <= (size_t)lc || (q = memchr(s, ' ', sz)) == NULL) {
			if(l < (la + sz + 1)) {
				l = la + sz + 100;
				space = xrealloc(space, l);
			}
			memcpy(space + la, s, sz);
			la += sz;
			space[la++] = '\n';
			continue;
		}
		/*   */
		while(1) {
			size_t sz2 = (q - s);
			if(mbchar_len(s, sz2) >= (size_t)lc ||
			  (t = memchr(q + 1, ' ', sz - sz2 - 1)) == NULL ||
			  mbchar_len(s, (t-s)) > (size_t)lc) {
				x = q + 1;
				while(x < e && *x == ' ')
					x++;
				if(x < e && *x == '\n')
					x++;
				if(l < (la + sz2 + 1)) {
					l = la + sz2 + 100;
					space = xrealloc(space, l);
				}
				memcpy(space + la, s, sz2);
				la += sz2;
				space[la++] = '\n';
				break;
			}
			q = t;
		}
	}
	free(x0);
	space = xrealloc(space, la);
}

static void T(struct REPLACES *rs)
{
    char outstr[200];
    time_t t;
    struct tm *tmp;
    char *x, *o;

    time(&t);
    tmp = localtime(&t);
    if (tmp == NULL)
	error_msg(3, "localtime: %m");

    x = set_repl(rs, &la, NULL);
    x = xrealloc(x, la + 1);
    x[la] = '\0';
    if(la > sizeof(outstr)) {
	o = xmalloc(la *= 2);
    } else {
	o = outstr;
	la = sizeof(outstr);
    }
    la = strftime(o, la, x, tmp);
    if (la != 0) {
	space = xrealloc(space, la);
	memcpy(space, o, la);
    } else {
	free(space);
	space = NULL;
	la = 0;
    }
    if(o != outstr)
	free(o);
    free(x);
}

static unsigned int replace1c(int f, unsigned int c)
{
	struct REPLACES *rs;

	replace1c_buf(f, c, m0, idx);
	for(rs = replaces; rs != NULL; rs = rs->next)
		if(rs->lf)
			replace1c_buf(f, c, rs->find, rs->lf);
	return c;
}

static void err_find_free_bytes(char c) __attribute__ ((__noreturn__));
static void err_find_free_bytes(char c)
{
	error_msg(3, "     \\%c", c);
}


static void new_input(const char *name, char *isuffix)
{
	unsigned int ch, c1 = 0, cu = 0, c1_old = 0, cu_old = 0, nf = 0;
	struct REPLACES *rs;

	if(input_file != NULL) {
		free(m0);
		allow_replace &= ~BIT_INPUT_STORED;
		nf = 1;
	}
	memcpy(used, uused, sizeof(used));
	bufsz = idx = 0;
	m0 = load_file(name, NULL, "", isuffix);
	if(isuffix != NULL) {
		if(isuffix[0] != '\0') {
			char *bak = NULL;
			if(asprintf(&bak, "%s%s", input_file, isuffix) < 0 || bak == NULL)
				error_msg(1, "       '%s': %m", input_file);
			if(rename(input_file, bak) < 0)
				error_msg(1, "   '%s'   '%s': %m", input_file, bak);
			free(bak);
		}
		sout = freopen(input_file, "w", sout);
		if(sout == NULL)
			error_msg(1, "  '%s': %m", input_file);
	}

	for(ch = 1; ch < 128; ch++) {
		if(used[ch]) continue;
		if(used[0] && c1 == 0)
			c1 = ch;
		else if(flg_need_unused && cu == 0)
			cu = ch;
		else
			break;
	}
	if(used[0]) {
		if(!c1)
			err_find_free_bytes('0');
		used[0] = replace1c(0, c1);
	}
	if(flg_need_unused) {
		if(cu == 0)
			err_find_free_bytes('y');
		for(rs = replaces; rs != NULL; rs = rs->next) {
			if((rs->mods & BIT_HAVE_UNUSED)) {
				size_t nr;
				char *find = rs->find;
				for(nr = 0; nr < rs->n_offs; nr++)
					if(rs->codes[nr] == CODE_UNUSED)
						find[rs->offs[nr]-1] = cu;
			}
		}
		used[1] = cu;
	} else {
		used[1] = 0;
	}
	for(rs = replaces; rs != NULL; rs = rs->next) {
		if((rs->mods & (BIT_REGEX|BIT_DYN_REGEX)) == BIT_REGEX) {
			if(nf == 1 && (c1_old != c1 || cu_old != cu))
				regfree(rs->R);
			if(nf == 0 || c1_old != c1 || cu_old != cu)
				xregcomp(rs->R, " ", rs, rs->find, rs->lf);
		}
	}
}

static void dyn_xregcomp(struct REPLACES *rs)
{
	if((rs->mods & BIT_DYN_REGEX)) {
		size_t l;
		char *x = set_repl(rs, &l, NULL);
		if(l == 0)
			error_msg(3, "%zu:       ", rs->line);
		xregcomp(rs->R, "  ", rs, x, l);
		free(x);
	}
}

static int find_pair_bracket(struct REPLACES *rs)
{
	size_t l;
	char *x = set_repl(rs, &l, NULL);

	if(l != 2) {
		char *x2, *p, *p2, *p0;
		size_t l1, l2, lx;

		x = deconvert(x, l, DECONVERT_NEED_ZTERM);
		if(mbchar_len(x, l) != 2)
			error_msg(1, "%zu:        ", rs->line);
		l1 = utf8_skip_len((unsigned char)*x);
		x2 = x + l1;
		l2 = utf8_skip_len((unsigned char)*x2);
		if((l1 + l2) != l)
			error_msg(1, "%zu:        utf-8 ", rs->line);
		if(l1 == l2 && !memcmp(x, x2, l1))
			error_msg(1, "%zu:        ", rs->line);
		p = memmem(m0, idx, x, l1);
		if(p == NULL) {
		 ret:
			free(x);
			return 0;
		}
		rm[0].rm_so = (p-m0);
		p0 = p + l1;
		l = 1;
		for(;;) {
			lx = idx - (p0-m0);
			p2 = memmem(p0, lx, x2, l2);
			if(p2 == NULL)
				goto ret;
			p = memmem(p0, lx, x, l1);
			if(p != NULL && p < p2) {
				l++;
				p2 = p + l1;
			} else {
				p2 += l2;
				if(!--l) break;
			}
			p0 = p2;
		}
		rm[0].rm_eo = (p2-m0);
	} else {
		char *cx, x2[2];

		x2[0] = *x;
		x2[1] = x[1];
		if(x[0] == x[1])
			error_msg(1, "%zu:        ", rs->line);
		free(x);
		l = 0;
		for(cx = m0; cx < (m0 + idx); cx++)
			if(*cx == x2[0]) {
				l = 1;
				break;
			}
		if(!l)
			return 0;
		rm[0].rm_so = (cx-m0);
		for(cx++; cx < (m0 + idx); cx++) {
			if(*cx == x2[0]) l++;
			else if(*cx == x2[1]) {
				if(!--l)
					break;
			}
		}
		rm[0].rm_eo = (cx-m0) + 1;
	}
	if(l)
		return 0;
	for(l = 1; l < MAX_OFFS; l++)
		rm[l].rm_so = rm[l].rm_eo = -1;
	return BIT_FOUND | BIT_INPUT_STORED;
}

static int execute(void)
{
	struct REPLACES *rs;
	size_t l;
	char *x;
	int eret = 0;

	for(rs = replaces; rs != NULL; rs = rs->next) {
		if((rs->mods & BIT_IS_RTEST) && (allow_replace & BIT_FOUND) == 0)
			continue;
		switch(rs->opt) {
		case 'j':
			rs = rs->find;
			break;
		case 't':
			if((allow_replace & BIT_FOUND))
				rs = rs->find;
			else {
				struct REPLACES *r = rs->find;
				if((rs->mods & BIT_SAVE_FLAGS)) {
					r->mods &= ~(BIT_COUNT_CMP_G|BIT_COUNT_CMP_L|BIT_ICONV_OK|BIT_FOUND);
					r->mods |= allow_replace;
				}
				if((rs->mods & BIT_IS_LOOP)) {
					r = r->next;
					allow_replace |= r->mods & BIT_FOUND;
					r->mods &= ~BIT_FOUND;
					r->mods |= BIT_IS_LOOP;
				}
			}
			break;
		case 'G':
			if((allow_replace & BIT_COUNT_CMP_G))
				rs = rs->find;
			else if((rs->mods & BIT_SAVE_FLAGS)) {
				struct REPLACES *r = rs->find;
				r->mods &= ~(BIT_COUNT_CMP_G|BIT_COUNT_CMP_L|BIT_ICONV_OK|BIT_FOUND);
				r->mods |= allow_replace;
			}
			break;
		case 'g':
			if((allow_replace & (BIT_FOUND|BIT_COUNT_CMP_G)))
				rs = rs->find;
			else if((rs->mods & BIT_SAVE_FLAGS)) {
				struct REPLACES *r = rs->find;
				r->mods &= ~(BIT_COUNT_CMP_G|BIT_COUNT_CMP_L|BIT_ICONV_OK|BIT_FOUND);
				r->mods |= allow_replace;
			}
			break;
		case 'L':
			if((allow_replace & BIT_COUNT_CMP_L))
				rs = rs->find;
			else if((rs->mods & BIT_SAVE_FLAGS)) {
				struct REPLACES *r = rs->find;
				r->mods &= ~(BIT_COUNT_CMP_G|BIT_COUNT_CMP_L|BIT_ICONV_OK|BIT_FOUND);
				r->mods |= allow_replace;
			}
			break;
		case 'l':
			if((allow_replace & (BIT_FOUND|BIT_COUNT_CMP_L)))
				rs = rs->find;
			else if((rs->mods & BIT_SAVE_FLAGS)) {
				struct REPLACES *r = rs->find;
				r->mods &= ~(BIT_COUNT_CMP_G|BIT_COUNT_CMP_L|BIT_ICONV_OK|BIT_FOUND);
				r->mods |= allow_replace;
			}
			break;
		case 'F':
			if((allow_replace & BIT_FOUND) == 0)
				rs = rs->find;
			else if((rs->mods & BIT_SAVE_FLAGS)) {
				struct REPLACES *r = rs->find;
				r->mods &= ~(BIT_COUNT_CMP_G|BIT_COUNT_CMP_L|BIT_ICONV_OK|BIT_FOUND);
				r->mods |= allow_replace;
			}
			break;
		case 'v':
			if((allow_replace & BIT_ICONV_OK))
				rs = rs->find;
			else if((rs->mods & BIT_SAVE_FLAGS)) {
				struct REPLACES *r = rs->find;
				r->mods &= ~(BIT_COUNT_CMP_G|BIT_COUNT_CMP_L|BIT_ICONV_OK|BIT_FOUND);
				r->mods |= allow_replace;
			}
			break;
		case '?':
			if((allow_replace & BIT_ICONV_OK) == 0)
				rs = rs->find;
			else if((rs->mods & BIT_SAVE_FLAGS)) {
				struct REPLACES *r = rs->find;
				r->mods &= ~(BIT_COUNT_CMP_G|BIT_COUNT_CMP_L|BIT_ICONV_OK|BIT_FOUND);
				r->mods |= allow_replace;
			}
			break;
		case ':':
			if((rs->mods & BIT_SAVE_FLAGS))
				allow_replace = rs->mods & (BIT_COUNT_CMP_G|BIT_COUNT_CMP_L|BIT_ICONV_OK|BIT_FOUND);
			break;
		case 'd':
			free(m0);
			m0 = NULL;
			idx = 0;
			allow_replace &= ~BIT_INPUT_STORED;
			break;
		case 'D':
			free(space);
			space = NULL;
			la = 0;
			break;
		case 'p':
			printb(sout, space, la, rs);
			break;
		case 'e':
			x = set_repl(rs, &l, NULL);
			printb(stderr, x, l, rs);
			free(x);
			eret = 1;
			break;
		case 'x':
			l = idx;
			idx = la;
			la = l;
			x = m0;
			m0 = space;
			space = x;
			allow_replace &= ~BIT_INPUT_STORED;
			break;
		case 'X':
			l = la;
			x = space;
			la = lb2;
			space = buf2;
			buf2 = x;
			lb2 = l;
			break;
		case 'Y':
			l = idx;
			idx = lb2;
			lb2 = l;
			x = m0;
			m0 = buf2;
			buf2 = x;
			allow_replace &= ~BIT_INPUT_STORED;
			break;
		case 'E':
			if(la == 0)
				error_msg(1, "%zu:     ", rs->line);
			x = deconvert(space, la, DECONVERT_NEED_ALLOC | DECONVERT_NEED_ZTERM);
			new_input(x, NULL);
			/* no free! */
			break;
		case 'w':
			x = set_repl(rs, &l, NULL);
			if(l == 0)
				error_msg(1, "%zu:    ", rs->line);
			else {
				FILE *out;

				x = deconvert(x, l, DECONVERT_NEED_ZTERM);
				out = fopen(x, ((rs->mods & BIT_H_MOD) ? "a" : "w"));
				if(out == NULL)
					error_msg(1, "%zu:      '%s': %m", rs->line, x);
				free(x);
				printb(out, space, la, rs);
				fclose(out);
			}
			break;
		case 'z':
			allow_replace = find_pair_bracket(rs);
			break;
		case 'N': case 'M': case 'n': case 'B':
			x = set_repl(rs, &l, NULL);
			x = deconvert(x, l, DECONVERT_NEED_ZTERM);
			if(rs->opt != 'n' && rs->opt != 'B') {
				char *e;
				long lc = count_strtol(x, &e, rs);
				if(*e == '+' || *e == '-' || *e == '*' || *e == '/' || *e == '&' || *e == '|') {
					long l2 = count_strtol(e + 1, NULL, rs);
					if(*e == '*')
						lc *= l2;
					else if(*e == '/')
						lc /= l2;
					else if(*e == '+')
						lc += l2;
					else if(*e == '&')
						lc &= l2;
					else if(*e == '|')
						lc |= l2;
					else
						lc -= l2;
					e = "";
				} else if(*e == ':') {
					e++;
					if((*e != 't' && *e != 'F' && *e != 'G' && *e != 'g' && *e != 'L' && *e != 'l')
							|| e[1] != '\0')
						error_msg(3, "%zu:   ", rs->line);
				} else if(*e != '\0')
					error_msg(3, "%zu:      ", rs->line);
				if(rs->opt == 'N') {
					count = lc;
				} else {
					long cmp = -1;
					if(lc == count) {
						allow_replace |= BIT_FOUND;
						if(*e == 't' || *e == 'g' || *e == 'l')
							cmp = 1;
						else if(*e != '\0')
							cmp = 0;
					} else {
						allow_replace &= ~BIT_FOUND;
						if(*e == 'F')
							cmp = 1;
					}
					if(count > lc) {
						allow_replace |= BIT_COUNT_CMP_G;
						if(*e == 'g' || *e == 'G')
							cmp = 1;
						else if(*e != '\0' && *e != 'F')
							cmp = 0;
					}
					if(count < lc) {
						allow_replace |= BIT_COUNT_CMP_L;
						if(*e == 'l' || *e == 'L')
							cmp = 1;
						else if(*e != '\0' && *e != 'F')
							cmp = 0;
					}
					if(cmp >= 0) count = cmp;
				}
			} else {
				if(rs->opt == 'n')
					count = (long)mbchar_len(x, l);
				else
					count = (long)l;
			}
			free(x);
			break;
		case 'R':
			dyn_xregcomp(rs);
			if(tre_regnexec(rs->R, m0, idx, MAX_OFFS, rm, 0) == 0)
				allow_replace = BIT_FOUND | BIT_INPUT_STORED;
			 else
				allow_replace = 0;
			if((rs->mods & BIT_IS_LOOP)) {
				rs->mods &= ~BIT_IS_LOOP;
				rs->mods |= allow_replace;
			}
			if((rs->mods & BIT_DYN_REGEX))
				regfree(rs->R);
			break;
		case 'c': case 'i': case 'a':
			set_repl(rs, &idx, &m0);
			allow_replace &= ~BIT_INPUT_STORED;
			break;
		case 'C': case 'A':
			set_repl(rs, &la, &space);
			if((rs->mods & BIT_CC_P_MOD))
				printb(sout, space, la, rs);
			break;
		case 'P':
			set_repl(rs, &lb2, &buf2);
			if((rs->mods & BIT_IS_PK))
				push_stack_level();
			break;
		case 'k':
			push_stack_level();
			break;
		case 'Z':
			m0 = xrealloc(m0, idx + lb2);
			memmove(m0 + lb2, m0, idx);
			memcpy(m0, buf2, lb2);
			idx += lb2;
			break;
		case 'K':
			pop_stack_level(rs->line, &buf2, &lb2);
			break;
		case 'S':
			pop_stack_level(rs->line, &m0, &idx);
			allow_replace &= ~BIT_INPUT_STORED;
			break;
		case 'O':
			x = set_repl(rs, &l, NULL);
			deconvert(x, l, 0);
			count = (long)mbchar_code(x, l);
			free(x);
			break;
		case 'V':
			x = set_repl(rs, &l, NULL);
			x = deconvert(x, l, DECONVERT_NEED_ZTERM);
			allow_replace &= ~(BIT_INPUT_STORED|BIT_ICONV_OK);
			{
				long chg = iconv_buf(x, rs->line, &m0, &idx);
				if(chg < 0) {
					count = -chg;
				} else {
					allow_replace |= BIT_ICONV_OK;
				}
			}
			free(x);
			break;
		case 'W':
			to_width(rs);
			break;
		case 'T':
			T(rs);
			break;
		case 'm': {
				char *e;
				x = set_repl(rs, &l, NULL);
				x = deconvert(x, l, DECONVERT_NEED_ZTERM);
				for(e = x; *e; e++)
					if(!( (*e >='0' && *e <= '9') ||
					       (*e >='A' && *e <= 'Z') ||
					       (*e >='a' && *e <= 'z') ||
					       *e == '_')) {
							l = 0;
							break;
						}
				if(!l)
					error_msg(3, "%zu: '%s':   ", rs->line, x);
				e = getenv(x);
				if(e == NULL)
					e = "";
				lb2 = strlen(e);
				buf2 = xrealloc(buf2, lb2);
				memcpy(buf2, e, lb2);
				free(x);
			}
			break;
		case 'Q':
			set_repl(rs, &idx, &m0);
			deconvert(m0, idx, DECONVERT_NEED_ZTERM);
			count = count_strtol(m0, NULL, rs);
			if(count < 0 || count > 255)
				error_msg(3, "%zu: '%ld':      0  255", rs->line, count);
			exit(count);
		/* unused [!-?JUbfhoqru{|~] */
		}
	}
	return eret;
}

int main(int argc, char **argv)
{
	int n, eret = 0;
	char *isuffix = NULL;

	initialize(argv[0]);
	while ((n = getopt(argc, argv, "e:f:i::")) > 0) {
		switch(n) {
		    case 'e':
			prg_buf = load_file(NULL, prg_buf, optarg, NULL);
			if(idx > 0 && prg_buf[idx-1] != '\n')
				prg_buf = load_file(NULL, prg_buf, "\n", NULL);
			break;
		    case 'f':
			prg_buf = load_file(optarg, prg_buf, NULL, NULL);
			if(idx > 0 && prg_buf[idx-1] != '\n')
				prg_buf = load_file(NULL, prg_buf, "\n", NULL);
			break;
		    case 'i':
			isuffix = optarg == NULL ? "" : optarg;
			break;
		    default:
			usage();
		}
	}
	if(prg_buf == NULL)
		usage();

	compile_script();

	argv += optind;
	if(*argv == NULL) {
		static char *argv_stdin[2] = { "-", NULL };
		if(isuffix != NULL)
			error_msg(2, "   ,    -i");
		argv = argv_stdin;
	}

	while(*argv != NULL) {
		new_input(*argv, isuffix);
		eret |= execute();
		argv++;
	}

	return eret;
}
