/*
 * Copyright (C) Vladimir Oleynik <dzo@simtreas.ru> 2015-2020
 * Licensed under GPLv2 or later, see file LICENSE in this tarball for details.
*/

#include "bsed.h"
#include <locale.h>
#include <ctype.h>
#include <wchar.h>
#include <wctype.h>
#include <iconv.h>

static int flg_utf8;
static char *encoding;
unsigned char used[256];

int up_low(const char *b, size_t bl, int flg_up_low, char *out, size_t need_convert, int chg_internally)
{
	size_t l;
	const unsigned char *p;
	const signed char *s, *e;
	mbstate_t ps, cps;
	wchar_t wc;

	if(need_convert == 0) {
		if(chg_internally != 0)
			memcpy(out, b, bl);
		return chg_internally;
	}
	if(flg_up_low == CODE_UPPER && !flg_utf8) {
		for(p = (const unsigned char *)b, l = bl; l != 0; p++, l--) {
			if(*p != used[0] && *p != used[1] && islower(*p))
				*out++ = toupper(*p);
			else
				*out++ = *p;
		}
		return 1;
	}
	if(flg_up_low == CODE_LOWER && !flg_utf8) {
		for(p = (const unsigned char *)b, l = bl; l != 0; p++, l--) {
			if(*p != used[0] && *p != used[1] && isupper(*p))
				*out++ = tolower(*p);
			else
				*out++ = *p;
		}
		return 1;
	}

	/* utf-8 */
	memset(&ps, 0, sizeof (mbstate_t));
	for(s = (const signed char *)b, e = s + bl; s < e; s++) {
		p = (const unsigned char *)s;
		if (*p == used[0] || *p == used[1]) {
			*out++ = *s;
		} else if (*s > 0) {
			if(flg_up_low == CODE_LOWER && isupper(*p))
				*out++ = tolower(*p);
			else if(flg_up_low == CODE_UPPER && islower(*p))
				*out++ = toupper(*p);
			else
				*out++ = *s;
		} else {
			char buf[MB_CUR_MAX + 1];
			size_t tmp = mbrtowc (&wc, (const char *)s, bl - (s - (signed char *)b), &ps);

			if ((size_t)(tmp) == (size_t)-1 || (size_t)(tmp) == (size_t)-2) {
				/* invalid bytes */
				*out++ = '?';
				/* reset states. */
				memset(&ps, 0, sizeof(mbstate_t));
			} else if (tmp == (size_t)0) {
				/* found '\0' char */
			} else {
				/* valid bytes */
				memset(&cps, 0, sizeof (mbstate_t));
				if(flg_up_low == CODE_UPPER && iswlower(wc)) {
					int ret = wcrtomb(buf, towupper(wc), &cps);
					if(ret > 0)
						out = memcpy(out, buf, ret) + ret;
				} else if(flg_up_low == CODE_LOWER && iswupper(wc)) {
					int ret = wcrtomb(buf, towlower(wc), &cps);
					if(ret > 0)
						out = memcpy(out, buf, ret) + ret;
				} else {
					out = memcpy(out, s, tmp) + tmp;
				}
				s += tmp - 1;
			}
		}
	}
	return 1;
}

size_t calc_up_low(const char *b, size_t bl, int flg_up_low, size_t *need_convert)
{
	size_t rez = 0, l;
	const unsigned char *p;
	const signed char *s, *e;
	mbstate_t ps, cps;
	wchar_t wc;

	if(flg_up_low == 0) {
		*need_convert = 0;
		return bl;
	}

	if(flg_up_low == CODE_UPPER && !flg_utf8) {
		p = (const unsigned char *)b;
		for(l = bl; l != 0; p++, l--) {
			if(*p != used[0] && *p != used[1] && islower(*p))
				rez++;
		}
		*need_convert = rez;
		return bl;
	}
	if(flg_up_low == CODE_LOWER && !flg_utf8) {
		p = (const unsigned char *)b;
		for(l = bl; l != 0; p++, l--) {
			if(*p != used[0] && *p != used[1] && isupper(*p))
				rez++;
		}
		*need_convert = rez;
		return bl;
	}

	/* utf-8 */
	memset(&ps, 0, sizeof (mbstate_t));
	l = 0;
	for(s = (const signed char *)b, e = s + bl; s < e; s++) {
		p = (const unsigned char *)s;
		if (*p == used[0] || *p == used[1]) {
			l++;
		} else if (*s > 0) {
			if(flg_up_low == CODE_LOWER && isupper(*p))
				rez++;
			else if(flg_up_low == CODE_UPPER && islower(*p))
				rez++;
			l++;
		} else {
			char buf[MB_CUR_MAX + 1];
			size_t tmp = mbrtowc (&wc, (const char *)s, bl - (s - (const signed char *)b), &ps);

			if ((size_t)(tmp) == (size_t)-1 || (size_t)(tmp) == (size_t)-2) {
				/* invalid bytes */
				l++;
				rez++;
				/* reset states. */
				memset(&ps, 0, sizeof(mbstate_t));
			} else if (tmp == (size_t)0) {
				/* found '\0' char */
				rez++;
			} else {
				/* valid bytes */
				s += tmp - 1;
				memset(&cps, 0, sizeof (mbstate_t));
				if(flg_up_low == CODE_UPPER && iswlower(wc)) {
					int ret = wcrtomb(buf, towupper(wc), &cps);
					if(ret > 0)
						l += ret;
					rez++;
				} else if(flg_up_low == CODE_LOWER && iswupper(wc)) {
					int ret = wcrtomb(buf, towlower(wc), &cps);
					if(ret > 0)
						l += ret;
					rez++;
				} else {
					l += tmp;
				}
			}
		}
	}
	*need_convert = rez;
	return rez == 0 ? bl : l;
}

void test_is_utf8(void)
{
  char *locale, *cp;

  setlocale(LC_ALL, "");

  /* Determine the current locale the same way as setlocale() does, according to POSIX. */
  locale = getenv("LC_ALL");
  if (!locale || !*locale) {
	locale = getenv("LC_CTYPE");
	if (!locale || !*locale)
		locale = getenv("LANG");
  }
  if (locale == NULL) {
	setlocale(LC_ALL, "ru_RU.KOI8-R");
	return;
  }
  if (locale && *locale) {
	/* The most general syntax of a locale (not all optional parts
	   recognized by all systems) is
	   language[_territory][.codeset][@modifier][+special][,[sponsor][_revision]]
	   To retrieve the codeset, search the first dot. Stop searching when
	   a '@' or '+' or ',' is encountered. */
	cp = locale;
	for (; *cp != '\0' && *cp != '@' && *cp != '+' && *cp != ','; cp++) {
		if (*cp == '.') {
			encoding = ++cp;
			for (; *cp != '\0' && *cp != '@' && *cp != '+' && *cp != ','; cp++);
			/* The encoding is now contained in the part from encoding to cp. */
			/* Check it for "UTF-8", which is the only official IANA name of
			UTF-8. Also check for the lowercase-no-dashes version, which is
			what some SystemV systems use. */
			if ((cp-encoding == 5 && !strncmp(encoding, "UTF-8", 5))
			 || (cp-encoding == 4 && !strncmp(encoding, "utf8", 4)))
				flg_utf8 = 1;
			break; /* not UTF-8 */
		}
	}
  }
  if(encoding != NULL) {
	size_t l = strlen(encoding) + 1;
	encoding = memcpy(xmalloc(l), encoding, l);
	cp = strpbrk(encoding, "@+,");
	if(cp != NULL)
		*cp = '\0';
  }
}

/* return number width chars */
size_t mbchar_len (const char *s, size_t l)
{
  size_t point = 0;
  size_t count = 0;
  mbstate_t ps;
  wchar_t wc;

  if(!flg_utf8)
	return l;

  memset(&ps, 0, sizeof (mbstate_t));
  while ((l - point) > 0) {
      size_t tmp = mbrtowc (&wc, s+point, l - point, &ps);
      if ((size_t)(tmp) == (size_t)-1 || (size_t)(tmp) == (size_t)-2) {
	  /* invalid bytes. asume a byte represents a character */
	  point++;
	  count++;
	  /* reset states. */
	  memset(&ps, 0, sizeof(mbstate_t));
      } else if (tmp == (size_t)0) {
	/* found '\0' char */
	break;
      } else {
	  /* valid bytes */
	  point += tmp;
	  count++;
      }
  }

  return count;
}

long mbchar_code (const char *s, size_t l)
{
  mbstate_t ps;
  wchar_t wc;

  if(s == NULL)
	return 0;
  if(!flg_utf8)
	return *(unsigned char *)s;
  memset(&ps, 0, sizeof (mbstate_t));
  if (l > 0 && s[0]) {
      size_t tmp = mbrtowc (&wc, s, l, &ps);
      if ((size_t)(tmp) == (size_t)-1 || (size_t)(tmp) == (size_t)-2) {
	  /* invalid bytes. asume a byte represents a character */
	  return (long)'?';
      } else if (tmp == (size_t)0) {
	/* found '\0' char */
	return 0;
      } else {
	  /* valid bytes */
	  return (long)wc;
      }
  }
  return 0;
}

long iconv_buf(char *ft, size_t line, char **pout, size_t *psz)
{
	char *from = NULL, *to;
	iconv_t cd;
	size_t f, t, t0;
	char *inb, *out;
	int from_is_utf8;
	size_t w1 = 0;

	to = strchr(ft, ':');
	if(to != NULL) {
		if(strchr(to + 1, ':') != NULL || to[1]=='\0')
			error_msg(3, "%zu:   V /:/", line);
		if(to != ft)
			from = ft;
		*to++ = '\0';
	} else {
		from = ft;
		if(*from == '\0')
			error_msg(3, "%zu:    V /:/", line);
		if(encoding == NULL)
			error_msg(3, "        V //");
		to = encoding;
	}
	if(from == NULL) {
		if(encoding == NULL)
			error_msg(3, "        V /:/");
		from = encoding;
	}
	from_is_utf8 = strcasecmp(from, "UTF8") == 0 || strcasecmp(from, "UTF-8") == 0;
	f = *psz;
	inb = *pout;
	t0 = t = *psz * 6;
	out = xmalloc(t + 1);

	cd = iconv_open(to, from);
	if( cd == (iconv_t)-1 )
		error_msg(3, "%zu:   V /%s:%s/", line, from, to);
	while(iconv(cd, &inb, &f, &out, &t) == (size_t)-1) {
		size_t s = from_is_utf8 ? utf8_skip_len(*(unsigned char *)inb) : 1;
		w1++;
		inb += s;
		f -= s;
		*out++ = '?';
		t--;
	}
	iconv_close(cd);
	t = t0 - t;
	out -= t;
	out = xrealloc(out, t + 1);
	out[t] = '\0';
	free(*pout);
	*pout = out;
	*psz = t;
	return -((long)w1);
}
